#############################################################################
# optim.py - Python extension module for optimization
#
# Author:  Paul Hughett
#
# Copyright (c) 2001 Trustees of the University of Pennsylvania
# See the file COPYING for license terms.
#
#
# This file contains various functions to support optimization in
# combination with the Pyvox module.
#
# $Id: optim.py,v 1.4 2001/05/18 21:10:22 hughett Exp $ Release $Name:  $
#############################################################################

import exceptions
import math
import pyvox, exim


class OptimizationError (exceptions.Exception) :
    def __init__(self, args=None) :
        self.args = args



#---------------------------------------------------------------------------
# obfunction - Objective function
#
# This base class defines an objective function for minimization, and also
# contains the various parameters that control the minimization.  The precise
# interpretation of the parameters is determined by the particular optimizer
# being used.  
#
# Attributes:
#
# npar is the number of dimensions in parameter space, or number of arguments
#    to the objective function.
#
# funct is the function to be minimized.  The user may elect to do strange
#    things by overriding the __call__ method, in which case the funct
#    attribute is traditionally set to None.
#
# xtol is the tolerance in the abscissa at which optimization will stop;
#    it is a vector of tolerances, one for each coordinate, or a plain
#     number for 1-D optimization.
#
# ftol is the tolerance in the objective function at which optimization will
#    stop.
#
# xopt is the best abscissa found so far.  Usually a vector, but a scalar for
#    1-D optimization.
#
# fopt is the best objective function value so far, or None if unknown.
#
# step is a vector of step sizes, or a scalar for 1-D optimization.
#
# dirset is a set of direction vectors; if both dirset and step are provided,
#    most optimizers will use dirset.  Not currently implemented.
#
# niter is the maximum number of iterations to run for; an iteration normally 
#    means one complete cycle through all coordinates.
#----------------------------------------------------------------------------

class obfunction :

    # Create an obfunction, with default parameter values
    def __init__(self, npar, funct=None) :
        self.npar  = npar
        self.funct = funct
        self.ftol  = 0.001
        self.fopt  = None
        self.niter = 25
        if npar > 1 :
            self.xtol  = npar * [0.001,]
            self.xopt  = npar * [0.0,]
            self.step  = npar * [1.0,]
        else :
            self.xtol = 0.001
            self.xopt = 0.0
            self.step = 1.0

    # Evaluate the function at a point
    def __call__(self, point) :
        return self.funct(point)



#------------------------------------------------------------------------
# univar - Univariate objective function derived from multivariate
#
# This class allows the user to create a univariate function of the
# form f1(x) = fn(p+xq) for some multivariate function fn and vectors
# p and q; minimization in a one-dimensional space is often used in
# minimization in multi-dimensions, and this class provides a tool
# for creating the function to be minimized in 1-d.  The attributes
# of the univariate objective function are set to default values under
# the assumption that p is the best x found so far and that q is a
# reasonable step size (i.e. size of the trust region in the indicated
# direction).
#
# FIXME: Add method to return xopt and fopt of original multivariate
# function?
#------------------------------------------------------------------------

class univar (obfunction) :

    def __init__(self, obfun, p, q) :
        obfunction.__init__(self, 1, obfun)
        self.p = p
        self.q = q
        self.xopt = 0.0
        self.step = 1.0
        self.xtol = 0.01
        self.ftol = obfun.ftol

    def __call__(self, x) :
        return self.funct(self.p+x*self.q)



#----------------------------------------------------------------------------
# golden - Golden section method for 1-d minimization
#
# This function implements the golden section method for one-dimensional
# optimization.  The argument is the univariate objective function to be 
# optimized  The value returned is the best abcissa found.
# Note that tol should not be set to a relative error of less than
# sqrt(eps), since numerical roundoff will pretty much guarantee that
# no better tolerance can be expected in general.
#
# FIXME: Add inverse parabolic interpolation for faster convergence?
#
# FIXME: First steps of bracket reduction may repeat evaluation of 
# function values computed during bracketing.
#
# FIXME: If step goes to zero (e.g. dir in powell goes to zero), this
# function will fail with an irrelevant error (unable to bracket minimum);
# it should perhaps terminate normally instead.
#----------------------------------------------------------------------------

def golden (obfun) :

    # Set up some parameters
    phi = (1 + math.sqrt(5)) / 2 
    maxsr = 10           # Max iterations to find initial bracket

    # Extract parameters from the objective function
    niter = obfun.niter
    xtol  = obfun.xtol
    ftol  = obfun.ftol

    # Obtain the initial bracket from the last optimum and the step
    a = obfun.xopt
    b = obfun.step + a

    # Order initial bracket to go downhill
    fa = obfun(a)
    fb = obfun(b)
    if fa < fb :
        (a, b) = (b, a)
        (fa, fb) = (fb, fa)

    # Find an initial bracket of three points
    c  = None
    for i in range(maxsr) :
        # print (a,b)         # DEBUG
        x = b + phi*(b-a)
        fx = obfun(x)
        if fb < fx :
            (a, b, c) = (a, b, x)
            (fa, fb, fc) = (fa, fb, fx)
            break
        else :
            (a, b) = (b, x)
            (fa, fb) = (fb, fx)
    if not c :
        raise OptimizationError, "Failed to bracket a minimum"

    # Reduce the bracket until tolerance is reached
    conv = None
    for i in range(niter) :
        # print (a,b,c)          # DEBUG

        if abs(c-a) < xtol :
            conv = 1
            break

        if fa - fb < ftol and fc - fb < ftol :
            conv = 1
            break
        
        if abs(c-b) < abs(b-a) :
            x = a + (b-a)/phi
            fx = obfun(x)
            if fx < fb :
                (a, b, c) = (a, x, b)
                (fa, fb, fc) = (fa, fx, fb)
            else :
                (a, b, c) = (x, b, c)
                (fa, fb, fc) = (fx, fb, fc)
        else :
            x = c + (b-c)/phi
            fx = obfun(x)
            if fx < fb :
                (a, b, c) = (b, x, c)
                (fa, fb, fc) = (fb, fx, fc)
            else :
                (a, b, c) = (a, b, x)
                (fa, fb, fc) = (fa, fb, fx)
    if not conv :
        raise OptimizationError, "Failed to converge to desired tolerance"

    # Save the middle of the final bracket in the obfunction and return
    obfun.xopt = b
    obfun.fopt = fb



#------------------------------------------------------------------------
# simple - Simple multi-dimensional minimization
#
# This function implements a simple (read naive) algorithm for multi-
# dimensional minimization--that is, it simply minimizes once along
# each coordinate axis and returns the best point found.
#
# NOTE that this function was originally written as a step towards the
# presumably better powell method, and is no longer actively maintained.
# That is, use the powell method unless you have some reason to believe
# that it won't work; and in that case, be prepared to upgrade this method
# with the features that have already been added to powell.
#
# FIXME: Golden often recalculates a function value that is
# already known here.
#
# FIXME: The initial step size should probably decrease in later interations
# of the outer loop.
#
# FIXME: Stop when the objective function is no longer improving, or when
# the increments in x are smaller than xtol.
#------------------------------------------------------------------------

def simple(obfun) :
    npar  = obfun.npar
    guess = obfun.xopt
    guess = pyvox.check_vector(guess, npar)
    steps = obfun.step
    niter = obfun.niter
    for i in range(niter) :
        for j in range(npar) :
            print guess
            dir = pyvox.array(npar, exim.double, 0)
            dir[j] = steps[j]
            fun = univar(obfun, guess, dir)
            golden(fun)
            optx = fun.xopt
            guess = guess + optx*dir
    return guess



#-------------------------------------------------------------------------
# powell - Powell direction set method for minimization 
#
# This function implements the Powell direction set method for miminization.
# The argument is the objective function to be optimized; all the other
# parameters are stored in the objective function.
# Returns the best point found.
#
# FIXME: Check if a direction gets too close to zero and either stop or
# raise an exception.
#
# FIXME: Try the variation of throwing out the direction that is most 
# nearly parallel to xopt - xopt0.
#
# FIXME: The best so far is returned in the obfunction; consider returning
# a boolean for improvement rather than the optimum x.  This may require
# changing some existing code (e.g. in rstest) to match.
#--------------------------------------------------------------------------

def powell(obfun) :
    npar  = obfun.npar     # Dimension of parameter space
    xopt  = obfun.xopt     # Best x found so far
    fopt  = obfun.fopt     # Best f(x) found so far
    xopt = pyvox.check_vector(xopt, npar)
    steps = pyvox.check_vector(obfun.step, npar)
    niter = obfun.niter

    # Set initial directions along coordinate axes
    dirset = npar*[0,]
    for i in range(npar) :
        dir = pyvox.array(npar, exim.double, 0)
        dir[i] = steps[i]
        dirset[i] = dir

    # Do niter rounds of improvement; any new optima are stored in
    # the obfunction for return to the caller
    for n in range(niter) :
        xopt0 = xopt         # Best x at start of round
        fopt0 = fopt         # Best f(x) at start of round
        for i in range(npar) :
            print    # Mark start of new dimension
            dir = dirset[i]
            fun = univar(obfun, xopt, dir)
            golden(fun)
            x1 = fun.xopt
            xopt = xopt + x1*dir
            obfun.xopt = xopt
            fopt = fun.fopt
            obfun.fopt = fopt
            if abs(x1) < 1.0 and abs(x1) > 0.3 :
                dirset[i] = x1*dir       # FIXME: Does this work well?
            elif abs(x1) < 0.3 :
                dirset[i] = 0.3 * dir
        if fopt0 and fopt - fopt0 < obfun.ftol :
            break      # No improvement in last round, so quit
        converged = 1
        for i in range(npar) :
            if abs(xopt[i] - xopt0[i]) > obfun.xtol[i] :
                converged = 0
        if converged == 1 :
            break      # No significant change in params, so quit
        newdir = xopt - xopt0
        dirset = dirset[1:npar] + [newdir,]

    return xopt
